//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================
//---------------------------------------------------------------------------
// TaeRegExp.cpp - implementation file for regular expression search class.
//---------------------------------------------------------------------------
// NOTE: this is a class, *NOT* an installable component!
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TaeRegExp.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
// define the size of the OutVector used by the class (200 is, according to
// my understanding, the largest useful size)
//
const int OutVectorSize = 200;
//---------------------------------------------------------------------------
// pcre_free() is, for some reason, not linking so #define it as a work-around
//
#define pcre_free(obj) ::free(obj)
//---------------------------------------------------------------------------
// simple methods used to get/set the bits in the FFlags member.  all
// of these methods operate similiarly and are not described further.
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetAnchored()
{
	return FFlags & PCRE_ANCHORED;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetAnchored(bool value)
{
	if (value) FFlags |= PCRE_ANCHORED;
	else FFlags &= ~PCRE_ANCHORED;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetCaseless()
{
	return FFlags & PCRE_CASELESS;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetCaseless(bool value)
{
	if (value) FFlags |= PCRE_CASELESS;
	else FFlags &= ~PCRE_CASELESS;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetDollarEndOnly()
{
	return FFlags & PCRE_DOLLAR_ENDONLY;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetDollarEndOnly(bool value)
{
	if (value) FFlags |= PCRE_DOLLAR_ENDONLY;
	else FFlags &= ~PCRE_DOLLAR_ENDONLY;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetDotAll()
{
	return FFlags & PCRE_DOTALL;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetDotAll(bool value)
{
	if (value) FFlags |= PCRE_DOTALL;
	else FFlags &= ~PCRE_DOTALL;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetExtended()
{
	return FFlags & PCRE_EXTENDED;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetExtended(bool value)
{
	if (value) FFlags |= PCRE_EXTENDED;
	else FFlags &= ~PCRE_EXTENDED;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetExtra()
{
	return FFlags & PCRE_EXTRA;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetExtra(bool value)
{
	if (value) FFlags |= PCRE_EXTRA;
	else FFlags &= ~PCRE_EXTRA;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetMultiLine()
{
	return FFlags & PCRE_MULTILINE;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetMultiLine(bool value)
{
	if (value) FFlags |= PCRE_MULTILINE;
	else FFlags &= ~PCRE_MULTILINE;
}
//---------------------------------------------------------------------------
bool __fastcall TTaeRegExp::GetUngreedy()
{
	return FFlags & PCRE_UNGREEDY;
}
//---------------------------------------------------------------------------
void __fastcall TTaeRegExp::SetUngreedy(bool value)
{
	if (value) FFlags |= PCRE_UNGREEDY;
	else FFlags &= ~PCRE_UNGREEDY;
}
//---------------------------------------------------------------------------
// other property getter/setter methods
//---------------------------------------------------------------------------
AnsiString __fastcall TTaeRegExp::GetLastErrMsg()
{
	if (!FLastErrMsg) return "no error.";
	return AnsiString(FLastErrMsg);
}
//---------------------------------------------------------------------------
int __fastcall TTaeRegExp::GetMatchCount()
{
	if (!FPcre || !FOffsets) return 0;
	return FOffsets->size();
}
//---------------------------------------------------------------------------
int __fastcall TTaeRegExp::GetMatchLength(unsigned int index)
{
	if (index > FOffsets->size()) return -1;
	return (*FOffsets)[index].FLength;
}
//---------------------------------------------------------------------------
int __fastcall TTaeRegExp::GetMatchStart(unsigned int index)
{
	if (index > FOffsets->size()) return -1;
	return (*FOffsets)[index].FOffset;
}
//---------------------------------------------------------------------------
// TTaeRegExp methods
//---------------------------------------------------------------------------
// TTaeRegExp::TTaeRegExp() - constructor.  allocates and initializes
// member data.
//
__fastcall TTaeRegExp::TTaeRegExp()
{
	// initialize all class data members
	FFlags = 0;
	FPcre = 0;
	FPcreExtra = 0;
	FLastErrOffset = 0;
	FMemStream = 0;
	FOffsets = 0;
	FPcreTables = 0;
	FLastErrMsg = 0;

	// set default compile options
	FFlags = PCRE_CASELESS | PCRE_MULTILINE;

	// allocate class members
	FPcreTables = pcre_maketables();  // no error condition defined in doc
	FMemStream = new TMemoryStream;
	FOutVector = new int[OutVectorSize * 3];
	FOffsets = new OffsetsVector;
}
//---------------------------------------------------------------------------
// TTaeRegExp::~TTaeRegExp() - destructor.  frees allocated memory.
//
__fastcall TTaeRegExp::~TTaeRegExp()
{
	if (FPcre) pcre_free(FPcre);
	if (FPcreExtra) pcre_free(FPcreExtra);
	if (FPcreTables) pcre_free((void*)FPcreTables);
	if (FOutVector) delete[] FOutVector;
	if (FOffsets) delete FOffsets;
}
//---------------------------------------------------------------------------
// TTaeRegExp::Compile() - compiles a regular expression. on success, returns
// true; failure returns false.
//
bool __fastcall TTaeRegExp::Compile(AnsiString regExp, bool optimize)
{
	// if no regular expression, fail
	if (!regExp.Length()) return false;

	// free memory previously allocated and re-initialize other global variables
	if (FPcre) {
		pcre_free(FPcre);		// should use pcre_free() per doc but link fails
		FPcre = 0;
		}
	if (FPcreExtra) {
		pcre_free(FPcreExtra);	// should use pcre_free() per doc but link fails
		FPcreExtra = 0;
		}

	// clear errors
	FLastErrMsg = 0;
	FLastErrOffset = 0;

	// empty match table
	FOffsets->clear();

	// compile the regular expression; return false on error
	FPcre = pcre_compile(regExp.c_str(), FFlags, &FLastErrMsg, &FLastErrOffset,
		FPcreTables);
	if (!FPcre) return false;

	// optimize the expression; note that this simply generates a table of
	// characters that might start a matching string; it is, therefore, not
	// useful if the user entered a string that will accept all characters
	// as a starting character (e.g., ".*somestring" would any sequence of
	// characters that ended with "somestring"); return false on error
	if (optimize) {
		FPcreExtra = pcre_study(FPcre, 0, &FLastErrMsg);
		if (FLastErrMsg) {
			pcre_free(FPcre);
			FPcre = 0;
			return false;
			}
		}

	// everything went well
	return true;
}
//---------------------------------------------------------------------------
// TTaeRegExp::FindPrior() - searches the match table for the match
// immediately preceding the position passed in selPos.  returns the
// matching index on success or -1 (unsigned) on failure.
//
unsigned int __fastcall TTaeRegExp::FindPrior(int selPos)
{
	// determine where the caret is currently located; locate index before
	// current selection
	unsigned int iCurrNdx = FOffsets->size();
	for (--iCurrNdx; iCurrNdx != (unsigned int) -1; iCurrNdx--)
		if ((*FOffsets)[iCurrNdx].FOffset < selPos) break;

	// increment the current vector location and validate
	if (iCurrNdx != (unsigned int) -1) return iCurrNdx;

	// no match
	return (unsigned int) -1;
}
//---------------------------------------------------------------------------
// TTaeRegExp::FindNext() - searches the match table for the match
// immediately following the position passed in selPos.  returns the
// matching index on success or -1 (unsigned) on failure.
//
unsigned int __fastcall TTaeRegExp::FindNext(int selPos)
{
	// determine where the caret is currently located; locate index past
	// current selection
	unsigned int iCurrNdx = 0;
	unsigned int size = FOffsets->size();
	for (; iCurrNdx < size; iCurrNdx++)
		if ((*FOffsets)[iCurrNdx].FOffset >= selPos) break;

	// increment the current vector location and validate
	if (iCurrNdx < size) return iCurrNdx;

	// no match
	return (unsigned int) -1;
}
//---------------------------------------------------------------------------
// TTaeRegExp::ScanError() - sets the LastErrMsg property to match the
// error string reported by pcre_exec().
//
void __fastcall TTaeRegExp::ScanError(int result)
{
	// clear error values
	FLastErrMsg = 0;
	FLastErrOffset = 0;

	// if a simple "no match found", beep and return no error
	if (result == -1) return;

	static const char* const msg[] = {
		"Unexpected error.",
		"No matches found.",
		"pcre_exec() passed NULL unexpectedly.",
		"Unrecognized option was passed to pcre_exec().",
		"Bad Magic (compiled code corrupt).",
		"Unknown node in compiled pattern.",
		"Insufficient memory.  Try reducing back references.",
		"Unknown error."
		};
	const int msgSize = sizeof(msg) / sizeof(msg[0]);

	// convert the error value into an array offset
	int errIndex = ::abs(result);

	// if the array index is out of bounds, force to last entry in table
	if (errIndex >= msgSize) errIndex = msgSize - 1;

	// set error message
	FLastErrMsg = msg[errIndex];
}
//---------------------------------------------------------------------------
// TTaeRegExp::Scan() - builds a table of locations that match the text
// in the TRichEdit passed.  returns true on success or false on error.
// note that a simple failure to match any text is not an error -- if
// the caller needs to know if any matches were found, the caller can
// check the MatchCount property.
//
bool __fastcall TTaeRegExp::Scan(TRichEdit* richEdit)
{
	// if insufficient information to scan, return false
	if (!richEdit) return false;

	// clear error information
	FLastErrMsg = 0;
	FLastErrOffset = 0;

	// create a memory stream if not already created
	TMemoryStream* pMemStream = new TMemoryStream();

	// save current RE PlainText property
	bool plainText = richEdit->PlainText;

	// set PlainText to true to ensure that only plain text is streamed out
	richEdit->PlainText = true;

	// save modified state (SaveToStream() may reset?)
	bool modified = richEdit->Modified;

	// stream out the RE control's contents
	richEdit->Lines->SaveToStream(pMemStream);

	// restore modified state
	richEdit->Modified = modified;

	// restore PlainText property
	richEdit->PlainText = plainText;

	// search for all occurrances of the pattern, saving the locations
	// and lengths in the Offsets vector
	FOffsets->clear();
	int startPos = 0;
	int result = 0;
	int flags = 0;
#if 0 // old PCRE 2.x code
	while ((result = pcre_exec(FPcre, FPcreExtra,
		(const char*) pMemStream->Memory + startPos,
		pMemStream->Position - startPos, flags,
		FOutVector, OutVectorSize)) > 0) {

		// save the match location in the Offsets vector
		OffsetInfo info(FOutVector[0] + startPos, FOutVector[1] - FOutVector[0]);
		FOffsets->push_back(info);
#else	// PCRE 3.4 code
	while ((result = pcre_exec(FPcre, FPcreExtra,
		(const char*) pMemStream->Memory, pMemStream->Position, startPos,
		flags, FOutVector, OutVectorSize)) > 0) {

		// save the match location in the Offsets vector
		OffsetInfo info(FOutVector[0], FOutVector[1] - FOutVector[0]);
		FOffsets->push_back(info);
#endif

		// update the starting position
		startPos = info.FOffset + info.FLength;

		// if we are past eol, get out quickly
		if (startPos >= pMemStream->Position) {
			result = -1;
			break;
			}

		// if we are not at the beginning of a line (paragraph), then we need
		// to set the PCRE_NOTBOL flag
		const char c = *((const char *) pMemStream->Memory + startPos - 1);
		flags = 0;
		// if using RE 2.0+, change the following to:
		//		if (c != '\r') flags |= PCRE_NOTBOL;
		// for an explanation, see the "Circumflex and dollar" topic in the
		// RegExpSyntax.rtf document in the distribution file.
		if (c != '\n') flags |= PCRE_NOTBOL;
		}

	// delete stream (not needed any more)
	delete pMemStream;

	// if something other than no match, set error info and fail
	if (result != -1) {
		ScanError(result);
		return false;
		}

	// return success
	return true;
}

//---------------------------------------------------------------------------
// TTaeRegExp::GetVersion() returns the PCRE version
//
AnsiString __fastcall TTaeRegExp::GetVersion(void)
{
	return pcre_version();
}
//---------------------------------------------------------------------------
// end of TaeRegExp.cpp
//---------------------------------------------------------------------------
//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================

