//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================
//---------------------------------------------------------------------------
// TaeRegExp.h - header file for regular expression search class.
//---------------------------------------------------------------------------
// NOTE: this is a class, *NOT* an installable component!
//---------------------------------------------------------------------------
#ifndef TaeRegExpH
#define TaeRegExpH
#include <vcl\ComCtrls.hpp>
#include "pcre.h"
//---------------------------------------------------------------------------
// include STL vector header and define OffsetInfo class as a vector
//
#include <vector>

class OffsetInfo {
public:
	int FOffset;
	int FLength;
	OffsetInfo(int offset = 0, int length = 0) {
		FOffset = offset; FLength = length;
		};
};

typedef std::vector<OffsetInfo> OffsetsVector;
//---------------------------------------------------------------------------
class TTaeRegExp : public TPersistent  {
protected:
	int FFlags;
	pcre_extra* FPcreExtra;
	pcre* FPcre;
	int FLastErrOffset;
	TMemoryStream* FMemStream;
	OffsetsVector* FOffsets;
	const unsigned char* FPcreTables;
	int* FOutVector;
	const char* FLastErrMsg;

	bool __fastcall GetAnchored();
	void __fastcall SetAnchored(bool value);
	bool __fastcall GetCaseless();
	void __fastcall SetCaseless(bool value);
	bool __fastcall GetDollarEndOnly();
	void __fastcall SetDollarEndOnly(bool value);
	bool __fastcall GetDotAll();
	void __fastcall SetDotAll(bool value);
	bool __fastcall GetExtended();
	void __fastcall SetExtended(bool value);
	bool __fastcall GetExtra();
	void __fastcall SetExtra(bool value);
	bool __fastcall GetMultiLine();
	void __fastcall SetMultiLine(bool value);
	bool __fastcall GetUngreedy();
	void __fastcall SetUngreedy(bool value);

	AnsiString __fastcall GetLastErrMsg();
	int __fastcall GetMatchCount();
	int __fastcall GetMatchLength(unsigned int index);
	int __fastcall GetMatchStart(unsigned int index);
	void __fastcall ScanError(int result);

	AnsiString __fastcall GetVersion(void);

public:
	__fastcall TTaeRegExp();
	__fastcall ~TTaeRegExp();
	bool __fastcall Compile(AnsiString regExp, bool optimize = true);
	bool __fastcall Scan(TRichEdit* richEdit);
	unsigned int __fastcall FindPrior(int selPos);
	unsigned int __fastcall FindNext(int selPos);

__published:
	__property bool Anchored = { read = GetAnchored, write = SetAnchored };
	__property bool Caseless = { read = GetCaseless, write = SetCaseless };
	__property bool DollarEndOnly = { read = GetDollarEndOnly, write = SetDollarEndOnly };
	__property bool DotAll = { read = GetDotAll, write = SetDotAll };
	__property bool Extended = { read = GetExtended, write = SetExtended };
	__property bool Extra = { read = GetExtra, write = SetExtra };
	__property bool MultiLine = { read = GetMultiLine, write = SetMultiLine };
	__property bool Ungreedy = { read = GetUngreedy, write = SetUngreedy };
	__property int Flags = { read = FFlags, write = FFlags };

	__property AnsiString LastErrMsg = { read = GetLastErrMsg };
	__property int LastErrOffset = { read = FLastErrOffset };

	__property int MatchCount = { read = GetMatchCount };
	__property int MatchLength[unsigned int index] = { read = GetMatchLength };
	__property int MatchStart[unsigned int index] = { read = GetMatchStart };

	__property AnsiString PcreVersion = { read = GetVersion };
};
//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------
// end of TaeRegExp.h
//---------------------------------------------------------------------------
//===========================================================================
// Copyright  2000 Thin Air Enterprises and Robert Dunn.  All rights reserved.
// Free for non-commercial use.  Commercial use requires license agreement.
// See http://home.att.net/~robertdunn/Yacs.html for the most current version.
//===========================================================================

